#!/bin/python

import asyncio
import os
from discord import \
    PCMVolumeTransformer, FFmpegPCMAudio, VoiceChannel, VoiceClient
from typing import List
from urllib.parse import urlparse, parse_qs, ParseResult
from os import path, system
import subprocess
from gremlin.discord.timestamps import from_seconds, parse_timestamp, stringify, duration
from gremlin.discord.ytdl import YTDL


def wrap(
    filename: str,
    prefix: str,
    suffix: str
) -> str:
    """
        Adds a prefix to the beginning of the filename.
    """
    filePath, baseName = path.split(filename)
    name, ext = path.splitext(baseName)
    baseName = prefix + name + suffix + ext
    return path.join(filePath, baseName)


clip_prefix = 'clipped_'
class Audio(PCMVolumeTransformer):

    def __init__(self, source, *, data, volume=0.5):
        super().__init__(source, volume)

        self.data = data

        self.title = data.get('title')
        self.url = data.get('url')


    @classmethod
    async def from_url(
        cls,
        url: str,
        *,
        start: str = None,
        end: str = None,
        clip: List[float],
        loop: asyncio.AbstractEventLoop = None,
        stream: bool = False
    ):
        # Parse time stamps if there are any.
        startTime = parse_timestamp(start) if start else None
        endTime = parse_timestamp(end) if end else None

        # Check if a time stamp was specified with the URL.
        parsedUrl: ParseResult = urlparse(url)
        if parsedUrl.query:
            query = parse_qs(parsedUrl.query)

            if query and 't' in query:
                try:
                    startTime = from_seconds(int(query['t'][0]))
                except ValueError:
                    pass

        yt = await YTDL(
            url,
            'cache',
            loop=loop,
            stream=stream
        )

        # Clip the video if time stamps are specified.
        if startTime or endTime or (clip and len(clip) == 2):

            clippedFile = wrap(
                yt.filename,
                clip_prefix,
                (f'_s{start}' if start else '') + \
                (f'_e{end}' if end else '') + \
                (f'_c{str(clip[0])}_{str(clip[1])}' if clip else '')
            )

            if not path.exists(clippedFile):

                # Download the video audio before editing.
                yt.download()

                # If clipping range is specifed, override.
                if clip:
                    result = subprocess.run(
                        [
                            'ffprobe',
                            '-v', 'error',
                            '-show_entries', 'format=duration',
                            '-of', 'default=noprint_wrappers=1:nokey=1',
                            yt.filename
                        ],
                        stdout=subprocess.PIPE,
                        stderr=subprocess.STDOUT
                    )
                    pulledDuration = float(result.stdout)
                    startTime = from_seconds(clip[0] * pulledDuration)
                    endTime = from_seconds(clip[1] * pulledDuration)

                # Order matters for cutting speed.
                # https://stackoverflow.com/a/42827058/10167844
                before = f'-ss {stringify(startTime)}' if startTime else ''
                if endTime:
                    dur = duration(endTime, (startTime if startTime else from_seconds(0)))
                after = f'-to {stringify(dur)}' if endTime else ''

                system(f'ffmpeg -y -hide_banner -loglevel error {before} -i {yt.filename} -vn {after} -c copy {clippedFile}')

                # Remove the original file.
                os.remove(yt.filename)


            # Use the clipped file.
            filename = clippedFile

        # Handle unclipped files.
        else:

            # Download video audio if it wasn't previously downloaded.
            if not (path.exists(yt.filename)):
                yt.download()


        return cls(
            FFmpegPCMAudio(
                filename,
                options='-vn'
            ),
            data=yt.data
        )


    @staticmethod
    async def play(
        url: str,
        voice_clients: List[VoiceClient],
        voice_channel: VoiceChannel,
        loop: asyncio.AbstractEventLoop,
        *,
        start: str = None,
        end: str = None,
        clip: List[float] = None
    ):
        """
            Plays the audio of a YouTube video into a voice channel.
        """

        # Check if the client is already connected to the targetted voice channel.
        voiceClient = next(
            (v for v in voice_clients if v.channel == voice_channel),
            None
        )

        # Connect if not already connected.
        if not voiceClient:
            voiceClient = await voice_channel.connect()
            
        player = await Audio.from_url(
            url,
            start=start,
            end=end,
            clip=clip
        )

        # Play audio into the voice channel.
        voiceClient.play(
            player,
            after=lambda error: asyncio.run_coroutine_threadsafe(
                voiceClient.disconnect(),
                loop
            )
        )


from asyncio import AbstractEventLoop, get_event_loop
from typing import Union, Dict, Any

def ytdl_format_options(directory: str):
    return {
        'format': 'bestaudio/best',
        'outtmpl': f'{directory}/%(extractor)s-%(id)s-%(title)s.%(ext)s',
        'restrictfilenames': True,
        'noplaylist': True,
        'nocheckcertificate': True,
        'ignoreerrors': False,
        'logtostderr': False,
        'quiet': True,
        'no_warnings': True,
        'default_search': 'auto',
        'source_address': '0.0.0.0', # bind to ipv4 since ipv6 addresses cause issues sometimes
        'cachedir': False # https://stackoverflow.com/a/32105062/10167844
    }

class YTDL:

    def __init__(
        self,
        url: str,
        directory: str,
        *,
        use_ytdlp: bool = True,
        loop: AbstractEventLoop = None,
        stream: bool = None
    ) -> None:
        self._url = url
        self._directory = directory
        self._use_ytdlp = use_ytdlp
        self._loop = loop
        self._stream = stream


    @classmethod
    async def create(
        cls,
        url: str,
        directory: str,
        *,
        use_ytdlp: bool = True,
        loop: AbstractEventLoop = None,
        stream: bool = None
    ):
        self = YTDL(
            url,
            directory,
            use_ytdlp=use_ytdlp,
            loop=loop,
            stream=stream
        )

        await self.load()

        return self


    async def load(self):
        '''
            Extract info from video.
        '''
        if self._use_ytdlp:
            import yt_dlp as ytdl
        else:
            import youtube_dl as ytdl

        # Suppress noise about console usage from errors
        ytdl.utils.bug_reports_message = lambda: ''

        self._ytdl = ytdl.YoutubeDL(ytdl_format_options(self._directory))

        # Get the video info.
        self._loop = self._loop or get_event_loop()
        self._data = await self._loop.run_in_executor(
            None,
            lambda: self._ytdl.extract_info(
                self._url,
                download=False
            )
        )
        
        # If it's a playlist, grab the first item.
        if 'entries' in self._data:
            self._data = self._data['entries'][0]

        # Get the filename.
        self._filename = self._data['url'] \
            if self._stream \
            else self._ytdl.prepare_filename(self._data)


    def download(self) -> None:
        self._ytdl.download([self._url])

    @property
    def data(self) -> Union[Any, Dict[str, Any], None]:
        return self._data

    @property
    def filename(self) -> str:
        return self._filename


import json
import requests
from typing import Dict
from urllib.parse import quote
from datetime import datetime
from pytz import UTC


class InvalidTokenError(Exception):
    pass

class ExpiredTokenError(Exception):
    pass


class DiscordAuth:
    def __init__(
        self,
        discord_api: str,
        redirect_url: str,
        client_id: str
    ):
        super()
        self.redirect_url = redirect_url
        self.oath_api = f'{discord_api}/oauth2'
        self.client_id = client_id


    def request_authorization(
        self,
        state: str,
        scope: str = 'identity',
        prompt: str = 'none'
    ) -> str:
        """
            Request a URL the user can use to request authorization
            from Discord.
        """
        if not state:
            raise ValueError('A state is required.')

        redirect_url = quote(self.redirect_url, safe='')
        auth_url = f'{self.oath_api}/authorize?response_type=code' + \
            f'&client_id={self.client_id}&state={state}&scope={scope}' + \
            f'&redirect_uri={redirect_url}&prompt={prompt}'

        return auth_url


    def request_access(
        self,
        code: str,
        client_secret: str
    ) -> Dict:
        """
            Get the access token for a proper authorization.
        """

        if not code:
            raise ValueError('Auth code required.')

        if not client_secret:
            raise ValueError('Client secret required.')

        data = {
            'client_id': self.client_id,
            'client_secret': client_secret,
            'grant_type': 'authorization_code',
            'code': code,
            'redirect_uri': self.redirect_url
        }
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded'
        }
        response = requests.post(
            f'{self.oath_api}/token',
            data=data,
            headers=headers
        )

        try:
            # Check for errors.
            response.raise_for_status()

        except requests.HTTPError as e:

            # Print and send up.
            msg = f'Status: {response.status_code}\nReason: {response.reason}' \
                + f'Content: {response.content}'

            print(msg)

            raise e

        return json.loads(response.content)


    def refresh_access(
        self,
        refresh_token: str,
        client_secret: str
    ) -> Dict:
        if not refresh_token:
            raise ValueError('Refresh token required.')

        if not client_secret:
            raise ValueError('Client secret required.')

        data = {
            'client_id': self.client_id,
            'client_secret': client_secret,
            'grant_type': 'refresh_token',
            'refresh_token': refresh_token
        }
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded'
        }
        response = requests.post(
            f'{self.oath_api}/token',
            data=data,
            headers=headers
        )

        tokens = json.loads(response.content)
        return tokens


    def get_user(
        self,
        access_token: str
    ) -> Dict:
        """
            Pings Discord's API to get the user of the access token.
        """

        # Get user's auth info.
        response = requests.get(
            f'{self.oath_api}/@me',
            headers={
                'Authorization': f'Bearer {access_token}'
            }
        )

        # Throw exception if token isn't recognized by Discord.
        if not response.ok:
            raise InvalidTokenError()

        content = json.loads(response.content)

        now = datetime.utcnow().replace(tzinfo=UTC)
        expiry = datetime.fromisoformat(content['expires']).replace(tzinfo=UTC)

        # Check for expiration date and user.
        if expiry < now or 'user' not in content:
            raise ExpiredTokenError()

        return content['user']


import pytest
import pytest_asyncio
from gremlin.discord.ytdl import YTDL
from pathlib import Path
from shutil import rmtree


example_files = [
    ('https://www.youtube.com/watch?v=dQw4w9WgXcQ', 'youtube-dQw4w9WgXcQ-Rick_Astley_-_Never_Gonna_Give_You_Up_Official_Music_Video.webm'),  # rickroll
    ('https://www.youtube.com/watch?v=QXUSvSUsx80', 'youtube-QXUSvSUsx80-I_d_just_like_to_interject_for_a_moment_uwu.webm'),  # uwu linux
    ('https://www.youtube.com/watch?v=hFcLyDb6niA', 'youtube-hFcLyDb6niA-Todd_and_the_sweet_little_lies.webm'),  # todd
    ('https://www.youtube.com/watch?v=2rCP4CRRO7E', 'youtube-2rCP4CRRO7E-Gandalf_Europop_Nod.webm'),  # gandalf
    ('https://www.youtube.com/watch?v=dbn-QDttWqU', 'youtube-dbn-QDttWqU-_.webm'),  # peace and tranquility
    ('https://www.youtube.com/watch?v=YJg_2r09Ips', 'youtube-YJg_2r09Ips-Gura_reads_the_dummy_thicc_meme_in_an_Ara_Ara_voice....webm'),  # gura
]

example_data = [
    ('https://www.youtube.com/watch?v=dQw4w9WgXcQ'),  # rickroll
    ('https://www.youtube.com/watch?v=QXUSvSUsx80'),  # uwu linux
    ('https://www.youtube.com/watch?v=hFcLyDb6niA'),  # todd
    ('https://www.youtube.com/watch?v=2rCP4CRRO7E'),  # gandalf
    ('https://www.youtube.com/watch?v=dbn-QDttWqU'),  # peace and tranquility
    ('https://www.youtube.com/watch?v=YJg_2r09Ips'),  # gura
]

test_directory = 'test_cache'

@pytest.mark.asyncio
@pytest_asyncio.fixture
async def manage_directory():
    path = Path(test_directory)

    # Remove an existing directory.
    if path.exists():
        rmtree(path, ignore_errors=True)

    # Create a fresh directory.
    path.mkdir()

    # Run test.
    yield None

    # Remove directory.
    rmtree(path, ignore_errors=True)
    print('yest')


@pytest.mark.asyncio
@pytest.mark.timeout(2)
@pytest.mark.parametrize(
    'url, expected',
    example_files
)
async def test_filename(
    url: str,
    expected: str,
    manage_directory
):
    '''
        Filename should be available and formatted
        a certain way after info is retrieved from YT.
    '''
    ytdl = await YTDL.create(
        url,
        test_directory
    )

    expected_path = f'{test_directory}/{expected}'

    assert ytdl.filename == expected_path, f'Should be {expected_path}.'


@pytest.mark.asyncio
@pytest.mark.timeout(2)
@pytest.mark.parametrize(
    'url',
    example_data
)
async def test_data(
    url: str,
    manage_directory
):
    ytdl = await YTDL.create(
        url,
        test_directory
    )

    assert ytdl.data != None, 'Data expected.'


@pytest.mark.asyncio
@pytest.mark.timeout(10)
@pytest.mark.parametrize(
    'url, expected',
    example_files
)
async def test_download(
    url: str,
    expected: str,
    manage_directory
):
    ytdl = await YTDL.create(
        url,
        test_directory
    )

    ytdl.download()

    filepath = f'{test_directory}/{expected}'

    assert Path(filepath).exists(), f'File should be downloaded to {filepath}.' 


import pathlib
from setuptools import setup

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

setup(
   name='cutegremlin',
   version='1.1.6',
   description='A bunch of painfully dull helpers for AWS, Discord, and others resources.',
   license='MIT',
   long_description=README,
   long_description_content_type='text/markdown',
   url='https://github.com/JesseZhong/cutegremlin',
   author='Jesse',
   author_email='jessetakuto@gmail.com',
   packages=[
       'gremlin',
       'gremlin.aws',
       'gremlin.web',
       'gremlin.discord',
       'gremlin.db'
    ],
   install_requires=[
       'boto3',
       'requests',
       'discord.py',
       'python-dotenv',
       'youtube-dl',
       'yt-dlp',
       'lmdb'
    ]
)



from os import getenv, path, mkdir
from typing import Dict, List, Union
from dotenv import load_dotenv
from lmdb import open
import json


load_dotenv()
DB_DIR = getenv('DB_DIR', './db')

def open_db():
    if not path.exists(DB_DIR):
        mkdir(DB_DIR)

    dbenv = open(
        DB_DIR,
        create=True,
        max_dbs=0
    )

    return dbenv


def transact_get(
    transaction: any,
    key: str,
    *,
    default: Union[Dict, List] = {}
) -> Union[Dict, List]:
    return json.loads(
        transaction.get(
            key.encode(encoding='utf-8'),
            default=bytes(
                json.dumps(default),
                'utf-8'
            )
        )
    )

def get(
    key: str,
    dbenv: any = None,
    *,
    default: Union[Dict, List] = {}
) -> Union[Dict, List]:
    """
        Retrieves an item from db.
    """

    # Open the DB environment if no
    # previously opened instance was passed.
    env = dbenv if dbenv else open_db()

    result = None
    with env.begin() as trnx:
        result = transact_get(
            trnx,
            key,
            default=default
        )

    # Close a self-opened DB environment.
    if not dbenv:
        env.close()
        
    return result


def transact_update(
    transaction: any,
    key: str,
    item: Union[Dict, List]
):
    transaction.put(
        key.encode(encoding='utf-8'),
        bytes(json.dumps(item), 'utf-8')
    )


def update(
    key: str,
    item: Union[Dict, List],
    *,
    dbenv: any = None
):
    """
        Update an item in the db.
    """
    
    # Open the DB environment if no
    # previously opened instance was passed.
    env = dbenv if dbenv else open_db()

    with env.begin(write=True) as trnx:
        transact_update(
            trnx,
            key,
            item
        )

    # Close a self-opened DB environment.
    if not dbenv:
        env.close()